#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#define NDEBUG
#include <assert.h>

#include <cv.h>
#include <highgui.h>

#include "renju.h"

typedef struct line_9{
    int horizon[9];
    int vertical[9];
    int diagonal_1[9];
    int diagonal_2[9];
    int status;
}line_9_t;

typedef struct winpoint{
    int x[5];
    int y[5];
}winpoint_t;

static uchar* p_board_data; // 15x15 pixel
static char* p_board_data_text; // 15x15 pixel
static int p_board_width, p_board_height, p_board_size;
static CvMat p_board_mat; // cvMatrix header for p_borad_data
static IplImage* p_board_display; // resized borad image for display
static int p_board_scale; // board scale

static winpoint_t winpoint;

int renju_get_board_size()
{
    return p_board_size;
}
int renju_get_board_text_size()
{
    return p_board_size + p_board_height;
}
int renju_get_board_width()
{
    return p_board_width;
}
int renju_get_board_height()
{
    return p_board_height;
}
unsigned char* renju_get_board_data()
{
    return p_board_data;
}
char* renju_get_board_data_text()
{
    int i, x, y;
    for(i = y = 0; y < p_board_height; y++, i++)
    {
        for(x = 0; x < p_board_width; x++, i++)
        {
            int data = CV_MAT_ELEM( p_board_mat, uchar, y, x );
            if(data == EMPTY_GRID)
                p_board_data_text[i] = '+';
            else if(data == PLAYER0_STONE)
                p_board_data_text[i] = '@';
            else
                p_board_data_text[i] = 'o';
        }
        p_board_data_text[i] = '\n';
    }
    return p_board_data_text;
}
void renju_display_borad(int delay)
{
//    cvResize(&p_board_mat, p_board_display, CV_INTER_NN);
    cvShowImage( "board", p_board_display );
    cvWaitKey(delay);
}
void renju_save_borad()
{
//    cvResize(&p_board_mat, p_board_display, CV_INTER_NN);
    cvSaveImage("result.png", p_board_display);
}
static void initialize_board_image(char* board_name)
{
    int i;
    CvPoint pt1, pt2;
    if( board_name )
    {
        IplImage* texture = cvLoadImage(board_name);
        cvResize(texture, p_board_display, CV_INTER_LINEAR);
    }
    else
    {
        pt1.x = pt1.y = 0;
        pt2.x = (p_board_width+1)*2*p_board_scale;
        pt2.y = (p_board_height+1)*2*p_board_scale;
        cvRectangle(p_board_display, pt1, pt2, CV_RGB(153,102,0), -1, 8, 0 );
    }
    for(i = 0; i < p_board_width; i++)
    {
        pt1.x = pt2.x = (i+1)*2*p_board_scale;
        pt1.y = 2*p_board_scale;
        pt2.y = p_board_height*2*p_board_scale;
        cvLine(p_board_display, pt1, pt2, CV_RGB(0, 0, 0), 1, 8, 0 );
        pt1.y = pt2.y = pt1.x;
        pt1.x = 2*p_board_scale;
        pt2.x = p_board_width*2*p_board_scale;
        cvLine(p_board_display, pt1, pt2, CV_RGB(0, 0, 0), 1, 8, 0 );
    }
    pt1.x = 8*p_board_scale;
    pt1.y = 8*p_board_scale;
    cvCircle(p_board_display, pt1, (int)(0.3*p_board_scale), CV_RGB(0, 0, 0), -1, 8, 0 );
    pt1.y = 24*p_board_scale;
    cvCircle(p_board_display, pt1, (int)(0.3*p_board_scale), CV_RGB(0, 0, 0), -1, 8, 0 );
    pt1.x = 24*p_board_scale;
    cvCircle(p_board_display, pt1, (int)(0.3*p_board_scale), CV_RGB(0, 0, 0), -1, 8, 0 );
    pt1.y = 8*p_board_scale;
    cvCircle(p_board_display, pt1, (int)(0.3*p_board_scale), CV_RGB(0, 0, 0), -1, 8, 0 );
    pt1.x = 16*p_board_scale;
    pt1.y = 16*p_board_scale;
    cvCircle(p_board_display, pt1, (int)(0.3*p_board_scale), CV_RGB(0, 0, 0), -1, 8, 0 );
}
void renju_initialize_board(int width, int height, int scale, char* board_name)
{
    p_board_width = width;
    p_board_height = height;
    p_board_size = p_board_height*p_board_width;
    p_board_data = (uchar*)malloc(sizeof(uchar)*p_board_size);
    p_board_data_text = (char*)malloc(sizeof(char)*(p_board_size+p_board_height));
    p_board_mat = cvMat(p_board_height, p_board_width, CV_8UC1, (void*)p_board_data);
    p_board_display = cvCreateImage(cvSize(2*scale*(p_board_width+1),2*scale*(p_board_height+1)),
                                    IPL_DEPTH_8U, 3);
    memset(p_board_data, EMPTY_GRID, sizeof(uchar)*p_board_size);
    cvNamedWindow( "board", CV_WINDOW_AUTOSIZE );
//    printf("%d %d %d\n", p_board_mat.cols, p_board_mat.rows, p_board_mat.step);

    p_board_scale = scale;
    initialize_board_image(board_name);
}
static line_9_t get_line_9(int x, int y)
{
    line_9_t line_9;
    /* horizon */
    int i;
    for(i = -4;i < 5;i++)
    {
	if(x + i < 0 || x + i > p_board_width - 1)
	    line_9.horizon[i + 4] = 63;
	else
	    line_9.horizon[i + 4] = CV_MAT_ELEM( p_board_mat, uchar, y, x + i );
    }
    /* vertical */
    for(i = -4;i < 5;i++){
	if(y + i < 0 || y + i > p_board_height - 1)
	    line_9.vertical[i + 4] = 63;
	else
	    line_9.vertical[i + 4] = CV_MAT_ELEM( p_board_mat, uchar, y + i, x );
    }
    /* diagonal_1 */
    for(i = -4;i < 5;i++){
	if(x + i < 0 || y + i < 0 || x + i > p_board_width - 1 || y + i > p_board_height - 1)
	    line_9.diagonal_1[i + 4] = 63;
	else
	    line_9.diagonal_1[i + 4] = CV_MAT_ELEM( p_board_mat, uchar, y + i, x + i );
    }
    /* diagonal_2 */
    for(i = -4;i < 5;i++){
	if(x - i < 0 || y + i < 0 || x - i > p_board_width - 1 || y + i > p_board_height - 1)
	    line_9.diagonal_2[i + 4] = 63;
	else
	    line_9.diagonal_2[i + 4] = CV_MAT_ELEM( p_board_mat, uchar, y + i, x - i );
    }
    return line_9;
}
static int check_chain5(int *line)
{
    int i = 0, chain = 0;
    while(1)
    {
	if((line[i] == PLAYER0_STONE || line[i] == PLAYER1_STONE) && line[i] == line[i + 1])
	    chain++;
	else
	    chain = 0;
	if(chain == 4)
	    return i;
	if(i == 7)
	    return 0;
	else
	    i++;
    }
}
static int check_chain3(int *line)
{
    int i,j;
    for(j=0;j<4;j++){
	if((line[j+1] == 0) && (line[j+5] == 0) && (line[j+2] == line[j+3]) && (line[j+2] == line[j+4]) && (line[j+2] != 0)){
	    for(i=1;i<6;i++)
		printf("%d ",line[j+i]);
	    printf("\n");
	    return 1;
	}
	else if((line[j+1] == line[j+4]) && (line[j+1] != 0) && (line[j] == 0) && (line[j+5] == 0) && (line[j+2] == line[j+1] || line[j+3] == line[j+1]) && (line[j+2] & line[j+3] == 0))
	    return 1;
    }
    return 0;
}
static int check_33(line_9_t *line_9, int x, int y)
{
    int ch3hor = check_chain3(line_9->horizon);
    int ch3ver = check_chain3(line_9->vertical);
    int ch3dia1 = check_chain3(line_9->diagonal_1);
    int ch3dia2 = check_chain3(line_9->diagonal_2);

    if(ch3hor+ch3ver+ch3dia1+ch3dia2 > 1)
	return 1;
    else
	return 0;
}
static int check_pattern(int x, int y)
{
    int i;
    line_9_t line_9 = get_line_9(x, y);
/*     if((check_chain5(line_9.horizon) == 1)||(check_chain5(line_9.vertical) == 1)|| */
/*        (check_chain5(line_9.diagonal_1) == 1)||(check_chain5(line_9.diagonal_2) == 1)) */
/* 	return WINNER; */
    int ch5hor = check_chain5(line_9.horizon);
    int ch5ver = check_chain5(line_9.vertical);
    int ch5dia1 = check_chain5(line_9.diagonal_1);
    int ch5dia2 = check_chain5(line_9.diagonal_2);
    if(ch5hor != 0){
	for(i=0;i<5;i++){
	    winpoint.x[i] = x + ch5hor - 7 + i;
	    winpoint.y[i] = y;
	}
	return WINNER;
    }
    else if(ch5ver != 0){
	for(i=0;i<5;i++){
	    winpoint.x[i] = x;
	    winpoint.y[i] = y + ch5ver - 7 + i;
	}
	return WINNER;
    }
    else if(ch5dia1 != 0){
	for(i=0;i<5;i++){
	    winpoint.x[i] = x + ch5dia1 - 7 + i;
	    winpoint.y[i] = y + ch5dia1 - 7 + i;
	}
	return WINNER;
    }
    else if(ch5dia2 != 0){
	for(i=0;i<5;i++){
	    winpoint.x[i] = x - (ch5dia2 - 7) - i;
	    winpoint.y[i] = y + ch5dia2 - 7 + i;
	}
	return WINNER;
    }
//    line_7_t line_7 = get_line_7(x, y);
    if(check_33(&line_9, x, y) == 1)
	return PUT_FOUL;
    else
	return PUT_SUCCESS;
}
inline void draw_stone(int x, int y, int stone)
{
    CvPoint pt;
    pt.x = (x+1)*2*p_board_scale;
    pt.y = (y+1)*2*p_board_scale;
    cvCircle(p_board_display, pt, (int)(0.9*p_board_scale), CV_RGB(stone, stone, stone), -1, 8, 0 );
}
inline int in_board(int x, int y)
{
    return ((0 <= x) && (x < p_board_width) && (0 <= y) && (y < p_board_height));
}
void renju_draw_winpoint(int stone)
{
    CvPoint pt;
    int i, j;
    CvScalar orig_color = cvScalarAll(stone*255);
    CvScalar next_color = orig_color;
    for(j = 0; ;j++){
	for(i=0;i<5;i++){
	    pt.x = (winpoint.x[i]+1)*2*p_board_scale;
	    pt.y = (winpoint.y[i]+1)*2*p_board_scale;
	    cvCircle(p_board_display, pt, (int)(0.9*p_board_scale), next_color, -1, 8, 0);
	}
	cvShowImage( "board", p_board_display );	
	if(cvWaitKey(250)!=-1)
	    break;
	if( j%2 )
	    next_color = orig_color;
	else
	    next_color = CV_RGB(255,0,0);
    }
}
int renju_put_stone(int x, int y, int player)
{
    int ret_code = NOP;
    int stone = PLAYER1_STONE;
    if(player == 0)
        stone = PLAYER0_STONE;
    int current = CV_MAT_ELEM( p_board_mat, uchar, y, x );
 
    if( current != EMPTY_GRID )
    {
        ret_code = PUT_FAILURE;
    }
    else
    {
        if(in_board(x, y))
        {
            CV_MAT_ELEM( p_board_mat, uchar, y, x ) = stone;
            ret_code = check_pattern(x, y);
            if((ret_code == PUT_SUCCESS)||(ret_code == WINNER))
	    {
		draw_stone(x, y, stone);
	    }
            else if(ret_code == PUT_FOUL)
            {
                CV_MAT_ELEM( p_board_mat, uchar, y, x ) = EMPTY_GRID;
            }
        }
        else
        {
            ret_code = PUT_OUTOF_BOARD;
        }
    }
    return ret_code;
}
